from graphql import get_introspection_query, build_client_schema, GraphQLObjectType
from graphql_query import Operation, Query, Argument, Variable, Field
from functools import cached_property
from contextlib import contextmanager
import requests
import time
import os
from .record_insights import RecordInsights
from pydantic import BaseModel

class TiloresAPI:
    """A simple API client to interact with a Tilores instance."""

    def __init__(self,
        *,
        api_url: str,
        token_url: str,
        client_id: str,
        client_secret: str,
        scope: list[str] = None
        ):
        self.api_url = api_url
        self.token_url = token_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.scope = scope or ["tilores/mutation.submit", "tilores/query.search", "tilores/query.entity"]
        self._access_token = None
        self._access_token_expires_at = None
        self._golden_records = {}

    @classmethod
    def from_environ(cls, **kwargs):
        return cls(
            api_url=os.environ['TILORES_API_URL'],
            token_url=os.environ['TILORES_TOKEN_URL'],
            client_id=os.environ['TILORES_CLIENT_ID'],
            client_secret=os.environ['TILORES_CLIENT_SECRET'],
            **kwargs
        )

    def fetch_access_token(self):
        response = requests.post(
            self.token_url,
            auth=(self.client_id, self.client_secret),
            data={'grant_type': 'client_credentials', 'scope': ' '.join(self.scope)},
        )
        response.raise_for_status()
        data = response.json()
        return (data['access_token'], data['expires_in'])

    @property
    def access_token(self):
        """Get the access token, refreshing it if necessary."""
        now = int(time.time())
        if self._access_token_expires_at and self._access_token_expires_at > now:
            return self._access_token
        else:
            access_token, expires_in = self.fetch_access_token()
            self._access_token = access_token
            self._access_token_expires_at = now + expires_in
            return self._access_token

    @cached_property
    def schema(self):
        result = self.gql(get_introspection_query(descriptions=True))
        return build_client_schema(result['data'])

    def gql(self, query, variables=None):
        """Perform a GraphQL query against the Tilores instance."""
        data = {'query': query}
        if variables is not None:
            def to_serializable(val):
                if isinstance(val, BaseModel):
                    return val.dict()  # Converts Pydantic model to dict
                elif isinstance(val, dict):
                    return {k: to_serializable(v) for k, v in val.items()}
                elif isinstance(val, list):
                    return [to_serializable(v) for v in val]
                return val
            data['variables'] = to_serializable(variables)
        response = requests.post(
            self.api_url,
            headers={
                "Authorization": f"Bearer {self.access_token}",
                "Content-Type": "application/json"
            },
            json=data
        )
        response.raise_for_status()
        return response.json()

    @cached_property
    def search_params(self):
        """Get a list of tuples of search parameter names and types for the search query."""
        search_params = []
        for name, graphql_input_field in self.schema.get_type('SearchParams').fields.items():
            search_params.append((name, graphql_input_field.type))
        return search_params

    @cached_property
    def search_param_names(self):
        """Get a list of search parameter names for the search query."""
        return [x for (x, _) in self.search_params]

    @cached_property
    def records_definition(self):
        """Returns the field definition for the records field."""
        def subfieldsOf(type):
            subfields = []
            for name, subfield in type.fields.items():
                if isinstance(subfield.type, GraphQLObjectType):
                    subfields.append(Field(name=name, fields=subfieldsOf(subfield.type)))
                else:
                    subfields.append(name)
            return subfields

        return Field(name='records', fields=subfieldsOf(self.schema.get_type('Record')))

    @cached_property
    def record_params(self, refresh=False):
        """Get a list of tuples of field names and their type for the RecordInput-type."""
        record_params = []
        for name, graphql_input_field in self.schema.get_type('RecordInput').fields.items():
            record_params.append((name, graphql_input_field.type))
        return record_params

    @cached_property
    def record_param_names(self):
        """Get a list of search parameter names for the search query."""
        return [x for (x, _) in self.record_params]

    def search(self, **params):
        """
        Perform a search query with the given parameters.

        See also: https://docs.tilotech.io/tilores/api/#query-search
        """
        [key for key in params.keys() if key != 'id']
        var_params = Variable(name='params', type='SearchParams!')
        operation = Operation(
            type='query',
            name='search',
            variables=[var_params],
            queries=[
                Query(
                    name='search',
                    arguments=[
                        Argument(name='input', value=Argument(name='parameters', value=var_params))
                    ],
                    fields=[
                        Field(name='entities', fields=[
                            'id',
                            'hits',
                            self.records_definition
                        ])
                    ]
                )
            ]
        )
        return self.gql(operation.render(), variables={'params': params})

    def entity_edges(self, entityID):
        """
        Searches for the edges of a single entity.
        """
        var_entity_id = Variable(name='entityID', type='ID!')
        operation = Operation(
            type='query',
            name='get_edges',
            variables=[var_entity_id],
            queries=[
                Query(
                    name='entity',
                    arguments=[
                        Argument(name='input', value=Argument(name='id', value=var_entity_id))
                    ],
                    fields=[
                        Field(name='entity', fields=[
                            'edges'
                        ])
                    ]
                )
            ]
        )
        return self.gql(operation.render(), variables={'entityID': entityID})


from typing import Optional, List
from graphql_query import Operation, Query, Argument, Variable, Field

DESC = 'DESC'
ASC = 'ASC'

class RecordInsights(Field):
    """
    This class implements a GraphQL query builder for RecordInsights queries on entities.

    See also: https://docs.tilotech.io/tilores/api/#record-insights
    """
    validate_field_names: Optional[List[str]] = None

    def __init__(self, alias=None, validate_field_names=None):
        super().__init__(name='recordInsights', alias=alias)
        self.validate_field_names = validate_field_names

    def assert_field_name(self, field_name):
        if self.validate_field_names is None:
            return
        assert field_name in self.validate_field_names, f'Field name {field_name} not present in field names: {self.validate_field_names!r}'

    def frequency_distribution(self, field_name, alias=None, top=1, direction=DESC):
        self.assert_field_name(field_name)
        alias = alias or field_name
        field = Field(
            name='frequencyDistribution',
            alias=alias,
            arguments=[
                Argument(name='field', value=f'"{field_name}"'),
                Argument(name='top', value=top),
                Argument(name='direction', value=direction),
            ],
            fields=[ 'value' ]
        )
        self.fields.append(field)
        return self

    def newest(self, sort_by_field_name, alias, fields):
        self.assert_field_name(sort_by_field_name)
        [self.assert_field_name(f) for f in fields if isinstance(f, str)]
        [self.assert_field_name(f.name) for f in fields if isinstance(f, Field)]
        field = Field(
            name='newest',
            alias=alias,
            arguments=[ Argument(name='field', value=f'"{sort_by_field_name}"') ],
            fields=fields
        )
        self.fields.append(field)
        return self

    def oldest(self, sort_by_field_name, alias, fields):
        self.assert_field_name(sort_by_field_name)
        [self.assert_field_name(f) for f in fields if isinstance(f, str)]
        [self.assert_field_name(f.name) for f in fields if isinstance(f, Field)]
        field = Field(
            name='oldest',
            alias=alias,
            arguments=[ Argument(name='field', value=f'"{sort_by_field_name}"') ],
            fields=fields
        )
        self.fields.append(field)
        return self

    def values(self, field_name, alias=None):
        self.assert_field_name(field_name)
        alias = alias or field_name
        field = Field(
            name='values',
            alias=alias,
            arguments=[ Argument(name='field', value=f'"{field_name}"')]
        )
        self.fields.append(field)
        return self

    def values_distinct(self, field_name, alias=None):
        self.assert_field_name(field_name)
        alias = alias or field_name
        field = Field(
            name='valuesDistinct',
            alias=alias,
            arguments=[ Argument(name='field', value=f'"{field_name}"')]
        )
        self.fields.append(field)
        return self




import graphql
import typing
from enum import Enum
import datetime
from pydantic import create_model, Field

TILORES_ROOT_TYPES = ['Record', 'RecordInput', 'SearchParams']

class PydanticFactory():
    def __init__(self, schema):
        self.references = {}
        self.schema = schema

    def generate(self):
        [self.create_model(model_name) for model_name in TILORES_ROOT_TYPES]
        return self.references

    def create_model(self, model_name:str):
        if model_name in self.references:
            return self.references[model_name]
        graphql_type = self.schema.get_type(model_name)
        assert graphql_type is not None, f'Cannot get type in schema for: {model_name!r}'
        fields = {}
        for field_name, field_type in graphql_type.fields.items():
            default_value, python_type = self.type_of(field_type)
            field = (python_type, Field(default_value, description=f'A value for {field_name} of type {python_type!r}'))
            fields[field_name] = field
        model = create_model(model_name, **fields, __doc__=graphql_type.description)
        self.references[model_name] = model
        return model

    def type_of(self, graphql_type, default_value=None):
        python_type = None
        match graphql_type:
            case graphql.type.definition.GraphQLInputField() | graphql.type.definition.GraphQLField():
                default_value, python_type = self.type_of(graphql_type.type, default_value)
            case graphql.type.definition.GraphQLNonNull():
                default_value = ...
                _, python_type = self.type_of(graphql_type.of_type, default_value)
            case graphql.type.definition.GraphQLScalarType(name='String') | graphql.type.definition.GraphQLScalarType(name='ID'):
                python_type = str
            case graphql.type.definition.GraphQLScalarType(name='Int'):
                python_type = int
            case graphql.type.definition.GraphQLScalarType(name='Boolean'):
                python_type = bool
            case graphql.type.definition.GraphQLScalarType(name='Float'):
                python_type = float
            case graphql.type.definition.GraphQLScalarType(name='Time'):
                python_type = datetime.time
            case graphql.type.definition.GraphQLScalarType(name='Date'):
                python_type = datetime.date
            case graphql.type.definition.GraphQLScalarType(name='DateTime'):
                python_type = datetime.datetime
            case graphql.type.definition.GraphQLList():
                inner_default_value, inner_python_type = self.type_of(graphql_type.of_type)
                if inner_default_value is not ...:
                    inner_python_type = typing.Optional[inner_python_type]
                python_type = list[inner_python_type]
            case graphql.type.definition.GraphQLEnumType(name=enum_model_name):
                if enum_model_name in self.references:
                    python_type = self.references[enum_model_name]
                else:
                    python_type = Enum(enum_model_name, [enum_field.value for enum_field in graphql_type.values.values()])
                    python_type.__doc__ = graphql_type.description
                    self.references[enum_model_name] = python_type
            case graphql.type.definition.GraphQLUnionType(name=union_model_name):
                inner_python_types = []
                for inner_graphql_type in graphql_type.types:
                    inner_python_type = self.create_model(inner_graphql_type.name)
                    inner_python_types.append(inner_python_type)
                python_type = typing.Union[*inner_python_types]
            case graphql.type.definition.GraphQLObjectType(name=model_name) | graphql.type.definition.GraphQLInterfaceType(name=model_name) | graphql.type.definition.GraphQLInputObjectType(name=model_name):
                python_type = self.create_model(model_name)
            case _:
                raise NotImplementedError(f'Unmatched case for GraphQL type: {graphql_type!r}')
        if default_value is not ...:
            python_type = typing.Optional[python_type]
        return default_value, python_type

